//---------------------------------
// 2019/9/16

#include "ArduCamManager.h"

#ifdef STORAGE_SPIFFS
#include <FS.h>
#else
#include <SPI.h>
#include "SdFat.h"
#endif


ArduCamManager::ArduCamManager(uint8_t cameraCS, uint8_t SDCS, uint8_t size, uint8_t quality){
  pinArduCamCS = cameraCS;
  pinSDCS = SDCS;
  jpegSize = size;
  jpegQuality = quality;
  arduCam = new ArduCAM(OV2640, cameraCS);

  if(statusNotConnected!=check()){
    init();
  }
}

ArduCamManager::~ArduCamManager(){
  delete arduCam;
}


// カメラモジュールのI2C, SPI接続を確認する
uint8_t ArduCamManager::check(){
  Wire.begin();
  SPI.begin();

  // I2C Sensor Check
  uint8_t pid = 0;
  uint8_t vid = 0;
  arduCam->wrSensorReg8_8(0xFF, 0x01);
  arduCam->rdSensorReg8_8(OV2640_CHIPID_HIGH, &vid);
  arduCam->rdSensorReg8_8(OV2640_CHIPID_LOW, &pid);
  if((vid != 0x26 ) && (( !pid != 0x41 ) || ( pid != 0x42 ))){
    return statusNotConnected;
  }

  // SPI Check
  uint8_t data = arduCam->read_reg(ARDUCHIP_TEST1);
  // Already initialized
  if(data == initData){
    return statusReady;
  }
  
  arduCam->write_reg(ARDUCHIP_TEST1, testData);
  data = arduCam->read_reg(ARDUCHIP_TEST1);
  if(data == testData){
    return statusNotInit;
  }
  return statusNotConnected;
}


// ArduCAM初期化とJPEGパラメータ設定
//  ARDUCHIP_TEST1レジスタにinitDataを書いておく このレジスタが設定済みであれば次回の初期化処理を飛ばす
//  初期化時はカメラが安定するまで4秒程度必要
void ArduCamManager::init(){
  arduCam->write_reg(ARDUCHIP_TEST1, initData);
  arduCam->set_format(JPEG);
  arduCam->InitCAM();
  setJpegParam(jpegSize, jpegQuality);
}


// JPEGのサイズと品質を設定
void ArduCamManager::setJpegParam(uint8_t size, uint8_t quality){
  arduCam->OV2640_set_JPEG_size(size);
  arduCam->wrSensorReg8_8(0xFF, 0x00);
  arduCam->wrSensorReg8_8(0x44, quality);
  jpegSize = size;
  jpegQuality = quality;
}


uint8_t ArduCamManager::getJpegSize(){
  return jpegSize;
}

uint8_t ArduCamManager::getJpegQuality(){
  return jpegQuality;
}


// 撮影を行ないストレージにJPEGファイルを書き込む
// 必要に応じて自動で初期化処理を入れる
bool ArduCamManager::takePhoto(String file){
  String path = storeDir + file;

  uint8_t stat=check();
  if(statusNotInit==stat){
    init();
    delay(4000);
  }else if(statusReady!=stat){
    return false;
  }

  arduCam->flush_fifo();
  arduCam->clear_fifo_flag();
  arduCam->start_capture();
  while(!arduCam->get_bit(ARDUCHIP_TRIG , CAP_DONE_MASK)){
    delay(1);
  }

  uint8_t*buf;
  buf = new uint8_t[bufferSize];
  uint8_t prev = 0;
  uint16_t index = 0;
    
  #ifdef STORAGE_SPIFFS
  if(SPIFFS.begin()){
    // Check free space
    struct FSInfo fsInfo;
    SPIFFS.info(fsInfo);
    if(fsInfo.totalBytes - fsInfo.usedBytes < (390*1024)){
      return false;
    }
    
    File file = SPIFFS.open(path, "w");
    if(file){
      arduCam->CS_LOW();
      arduCam->set_fifo_burst();
      #ifndef OV2640_MINI_2MP_PLUS
      buf[index] = SPI.transfer(0x0);
      #endif
      
      while(1){
        buf[index] = SPI.transfer(0x0);

        // End of JPEG file
        if(prev==0xFF && buf[index]==0xD9){
          arduCam->CS_HIGH();
          file.write(buf, index+1);
          break;
        }

        prev = buf[index];
        if(index==bufferSize-1){
          //arduCam->CS_HIGH();
          file.write(buf, bufferSize);
          //arduCam->CS_LOW();
          //arduCam->set_fifo_burst();
          index = 0;
        }else{
          index++;
        }
        delay(0);
      }
      file.close();
    }
    SPIFFS.end();
  }
  // Write to SD
  #else
  SdFat*SD = new SdFat;
  if(SD->begin(pinSDCS, SD_SCK_MHZ(spiClkSD))){
    if(!SD->exists(storeDir.c_str())){
      SD->mkdir(storeDir.c_str());
    }
    File file = SD->open(path, O_RDWR | O_CREAT);
    if(file){
      arduCam->CS_LOW();
      arduCam->set_fifo_burst();
      #ifndef OV2640_MINI_2MP_PLUS
      buf[index] = SPI.transfer(0x0);
      #endif
      
      while(1){
        buf[index] = SPI.transfer(0x0);

        // End of JPEG file
        if(prev==0xFF && buf[index]==0xD9){
          arduCam->CS_HIGH();
          file.write(buf, index+1);
          break;
        }

        prev = buf[index];
        if(index==bufferSize-1){
          arduCam->CS_HIGH();
          file.write(buf, bufferSize);
          arduCam->CS_LOW();
          arduCam->set_fifo_burst();
          index = 0;
        }else{
          index++;
        }
        delay(0);
      }
      file.close();
    }
  }
  delete SD;
  #endif
  delete[] buf;

  return true;
}
