#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <ESP8266mDNS.h>
#include <Wire.h>
#include <SPI.h>
#include <WiFiUdp.h>
#include <ArduCAM.h>
#include <ArduinoOTA.h>
#include "Storage.h"
#include "WebPageHandler.h"
#include "TimelyTaskManager.h"
#include "TSL2561I2C.h"
#include "ArduCamManager.h"


const String wifiSSID = "yourssid";      // お使いのWiFiのSSIDを入力してください
const String wifiPass = "yourpassword";  // お使いのWiFiのパスワードを入力してください

const uint8_t pinLED = 14;
const uint8_t pinSW = 0;
const uint8_t pinArduCamCS = 15;
const uint8_t pinSDCS = 2;
const uint8_t spiClkSD = 8;     // SD SPI clock MHz
const uint8_t pinSDA = 4;
const uint8_t pinSCL = 5;
const uint8_t pinMotion = 16;

// TimelyTaskManager定期処理用ID
const uint8_t tmidLuxLog = 0;



ESP8266WebServer server(80);
WebPageHandler*webPageHandler;       // Webページの処理を行っているクラス
TimelyTaskManager*timelyTaskManager; // インターネットから時刻の取得などを行っているクラス
ArduCamManager*arduCamManager;       // ArduCAM関係の処理を行っているクラス
TSL2561I2C tsl2561(pinSDA, pinSCL);  // 照度センサーの処理を行っているクラス

bool statWiFiConnected = true;       // WiFiがつながっているかの監視用
uint32_t lastPhoto = 0;              // 最後に写真を撮った時間を記憶
uint32_t lastStat = 0;               // 最後に定期処理をした時間を記憶


void setup() {
  Serial.begin(115200);
  delay(1);
  Storage::opLog("\n\n-----------------------\n");
  Storage::opLog("Program Start\n");

  // ピン設定
  pinMode(pinLED, OUTPUT);   // LED
  pinMode(pinSW, INPUT);     // Switch
  pinMode(pinMotion, INPUT_PULLDOWN_16);

  // SPI CS をHighにしておく
  pinMode(pinArduCamCS, OUTPUT);
  pinMode(pinSDCS, OUTPUT);
  digitalWrite(pinArduCamCS, HIGH);
  digitalWrite(pinSDCS, HIGH);
  
  // WiFi設定
  Storage::opLog("\nWiFi");
  WiFi.mode(WIFI_STA);
  WiFi.begin(wifiSSID.c_str(), wifiPass.c_str());
  
  while(WiFi.status()!=WL_CONNECTED){
    delay(100);
  }
  // WiFi.config(IPAddress(192, 168, 1, 30), WiFi.gatewayIP(), WiFi.subnetMask());  // IPを指定する場合はコメントをはずす
  Storage::opLog(" Connected IP : " + WiFi.localIP().toString() + "\n");


  // OTA
  ArduinoOTA.begin();
  

  // TimelyTaskManager
  timelyTaskManager = new TimelyTaskManager(3);
  timelyTaskManager->updateTime();
  Storage::opLog(timelyTaskManager->getTimeStr() + " TimelyTaskManager Started\n");  

  // ArduCam
  arduCamManager = new ArduCamManager(pinArduCamCS, pinSDCS, OV2640_800x600, ArduCamManager::qualityGood);
  arduCamManager->storeDir = "/arducam/";
  delay(2000);
  Storage::opLog(timelyTaskManager->getTimeStr() + " ArduCamManager Initialized\n");

  
  // WebPageHandler
  webPageHandler = new WebPageHandler(&server, arduCamManager, timelyTaskManager, webPageHandlerCallBack);

  
  // Web Server
  // 処理は全てwebPageHandlerに投げる
  server.onNotFound([](){
    webPageHandler->handleFile();
  });
  server.begin();
  
  Storage::opLog(timelyTaskManager->getTimeStr() + " Web Server Started\n");
}


void loop() {
  // モーションセンサー撮影処理
  pinMode(pinLED, OUTPUT);
  if(digitalRead(pinMotion)==1){
    if(webPageHandler->motionSensorStatus){
      if(millis() - lastPhoto > 30000){  // 最後の撮影から最低30秒空ける
        // 暗すぎる場合は撮影しない
        bool tslStat = tsl2561.meas();
        if((tslStat==false) || (tsl2561.lux>1)){
          Storage::opLog(timelyTaskManager->getTimeStr() + " Motion Photo");
          if(arduCamManager->takePhoto(timelyTaskManager->getTimeFileName(TimelyTaskManager::formatSecond)+".jpg")){
            Storage::opLog(" Success\n");
          }else{
            Storage::opLog(" Failed\n");
          }
          lastPhoto = millis();
          digitalWrite(pinLED,1);
        }
      }
    }
  }else{
    digitalWrite(pinLED,0);
  }
  
  server.handleClient();
  ArduinoOTA.handle();
  timelyTaskManager->updateTask();
  if(millis() > lastStat+3000){   // 3秒おきの処理
    lastStat = millis();
    stat();
  }
  delay(1);   // 消費電力低減
}

// 3秒おきに行う定期処理
void stat(){
  // WiFiの状態を記録
  if(WiFi.status()==WL_CONNECTED){
    // Changed from the last status
    if(!statWiFiConnected){
      statWiFiConnected = true;
      Storage::opLog(timelyTaskManager->getTimeStr() + " WiFi Connected\n" + WiFi.localIP().toString());
    }
  }else{
    if(statWiFiConnected){
      statWiFiConnected = false;
      Storage::opLog(timelyTaskManager->getTimeStr());
      Storage::opLog(" WiFi Disconnected\n");
    }
  }
}


// 照度ログ動作中に1分おきに実行する関数
void luxLog(){
  String timeStr = timelyTaskManager->getTimeStr();
  Storage::opLog(timeStr);
  Storage::opLog(" Lux log");

  TSL2561I2C tsl2561(pinSDA, pinSCL);
  if(!tsl2561.meas()){
    Storage::opLog(" TSL2561 Not Available\n");
    return;
  }
  String luxStr = tsl2561.getLuxStr();

  String path = String("/") + WebPageHandler::luxLogFileName;

  // Log on SPIFFS
  #ifdef STORAGE_SPIFFS
  if(SPIFFS.begin()){
    bool exists = SPIFFS.exists(path);
    File file = SPIFFS.open(path, "a");
    if(file){
      if(!exists){
        file.print("Time,Illuminance[lux]\n");
      }
      file.print(timeStr.c_str());
      file.print(",");
      file.print(luxStr.c_str());
      file.print("\n");
      file.close();
    }
    SPIFFS.end();
  }
  // Log on SD
  #else
  SdFat SD;
  if(SD.begin(pinSDCS, SD_SCK_MHZ(spiClkSD))){
    bool exists = SD.exists(path.c_str());
    File file = SD.open(path, FILE_WRITE);
    if(file){
      // Write header
      if(!exists){
        file.write("Time,Illuminance[lux]\n");
      }
      file.write(timeStr.c_str());
      file.write(",");
      file.write(luxStr.c_str());
      file.write("\n");
      file.close();
    }
  }
  #endif

  Storage::opLog(" Done\n");
}


// webPageHandlerから呼ばれるコールバック
// 照度ログの開始、停止ボタンが押されると呼び出される
void webPageHandlerCallBack(uint8_t action){
  switch(action){
  case WebPageHandler::luxLogStart:
    timelyTaskManager->setMinutelyTask(tmidLuxLog, 0, luxLog);
    Storage::opLog(timelyTaskManager->getTimeStr() + " Action Lux Log Started\n");
    break;
  case WebPageHandler::luxLogStop:
    timelyTaskManager->removeTask(tmidLuxLog);
    Storage::opLog(timelyTaskManager->getTimeStr() + " Action Lux Log Stopped\n");
    break;
  }
}
