//---------------------------------
// 2018/1/18

#include <Wire.h>
#include "TSL2561I2C.h"


TSL2561I2C::TSL2561I2C(uint8_t sda, uint8_t scl){
  pinSDA = sda;
  pinSCL = scl;
}


// Read 1Byte from register
uint8_t TSL2561I2C::readAddress(uint8_t addr){
  uint8_t c = addr | 0x80;
  Wire.begin(pinSDA, pinSCL);
  Wire.beginTransmission(i2cAddr);
  Wire.write(c);
  Wire.endTransmission();
  Wire.requestFrom(i2cAddr,1);
  return Wire.read();
}

// Read length Byte from register
void TSL2561I2C::readAddress(uint8_t addr, uint8_t*data, uint8_t length){
  uint8_t c = addr | 0x80;
  Wire.begin(pinSDA, pinSCL);
  Wire.beginTransmission(i2cAddr);
  Wire.write(c);
  Wire.endTransmission();
  Wire.requestFrom(i2cAddr,length);
  for(uint8_t i=0; i<length; i++){
    data[i] = Wire.read();
  }
}

// Write 1Byte to register
void TSL2561I2C::writeAddress(uint8_t addr, uint8_t data){
  uint8_t c = addr | 0x80;
  Wire.begin(pinSDA, pinSCL);
  Wire.beginTransmission(i2cAddr);
  Wire.write(c);
  Wire.write(data);
  Wire.endTransmission();
  return;
}


// Clear interrupt flag
void TSL2561I2C::clearInt(){
  Wire.begin(pinSDA, pinSCL);
  Wire.beginTransmission(i2cAddr);
  Wire.write(0xC0);
  Wire.endTransmission();
  return;
}

// ID Read
//  Return true if success
bool TSL2561I2C::idRead(){
  uint8_t id = readAddress(0xA);
  //Serial.print("ID : ");
  //Serial.println(id, HEX);
  id = id>>4;

  // TSL2561 ID
  if(id == 0x1 || id == 0x5){
    return true;
  }
  return false;
}


// Set gain and atime
void TSL2561I2C::timing(uint8_t again, uint8_t atime){
  writeAddress(0x1, (again<<4) | atime);
}

// ALS integration and get result to ch0 and ch1
void TSL2561I2C::alsIntegration(){
  uint8_t data[4];
  writeAddress(0x0, 0x0);  // Power OFF
  timing(again, atime);
  writeAddress(0x0, 0x3);  // Power ON
  switch(atime){
  case atime13ms:
    delay(20);
    break;
  case atime101ms:
    delay(120);
    break;
  case atime402ms:
    delay(430);
    break;
  }
  readAddress(0xC, data, 4);
  writeAddress(0x0, 0x0);  // Power OFF
  ch0 = data[1]<<8 | data[0];
  ch1 = data[3]<<8 | data[2];
}


// One time measurement and get lux
//  Automatically adjust gain and time
bool TSL2561I2C::meas(){
  if(idRead()==false){
    return false;
  }
  
  again = againHigh;
  atime = atime101ms;

  while(1){
    alsIntegration();
    if(ch0 > 37000 || ch1 > 37000){
      if(again == againHigh && atime == atime101ms){
        again = againLow;
      }else{
        break;
      }
    }else if(ch0 < 300 || ch1 < 300){
      if(again == againHigh && atime == atime101ms){
        atime = atime402ms;
      }else{
        break;
      }
    }else{
      break;
    }
  }
  calcLux();
  return true;
}


// Calculate lux
void TSL2561I2C::calcLux(){
  float fAtime;
  float fAgain;

  switch (atime){
  case atime13ms:
    fAtime = 13.7;
    break;
  case atime101ms:
    fAtime = 101;
    break;
  case atime402ms:
    fAtime = 402;
    break;
  }

  switch (again)
  {
  case againLow:
    fAgain = 1;
    break;
  case againHigh:
    fAgain = 16;
    break;
  }
  
  float ratio = (float)ch1 / (float)ch0;
  
  // T, FN, CL package
  if(ratio <= 0.5){
    lux = 0.0304*ch0 - (0.062*ch0 * pow(ratio,1.4));
  }else if(ratio <= 0.61){
    lux = 0.0224*ch0 - 0.031*ch1;
  }else if(ratio <= 0.8){
    lux = 0.0128*ch0 - 0.0153*ch1;
  }else if(ratio <=1.3){
    lux = 0.00146*ch0 - 0.00112*ch1;
  }else{
    lux = 0;
  }

  lux = lux * 16/fAgain * 402/fAtime;
}


// Set interrupt conditions
void TSL2561I2C::setInterrupt(bool enable, uint16_t lowTh, uint16_t highTh, uint8_t persist){
  writeAddress(0x0, 0x0);  // Power OFF
  writeAddress(0x6, persist&0xF);   // Stop interrupt
  writeAddress(0x2, lowTh&0xFF);
  writeAddress(0x3, lowTh>>8);
  writeAddress(0x4, highTh&0xFF);
  writeAddress(0x5, highTh>>8);
  
  if(enable){
    writeAddress(0x0, 0x3);  // Power ON
    writeAddress(0x6, 0x10 | (persist&0xF));  // Start interrupt
  }
}


String TSL2561I2C::getLuxStr(){
  char c[8];
  dtostrf(lux, 2, 1, c);
  return String(c);
}















